==> apache-beam-2.58.0-source-release.zip: beam-2.58.0/sdks/java/io/solace/src/main/java/org/apache/beam/sdk/io/solace/read/UnboundedSolaceReader.java <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.beam.sdk.io.solace.read;

import static org.apache.beam.vendor.guava.v32_1_2_jre.com.google.common.base.Preconditions.checkNotNull;

import com.solacesystems.jcsmp.BytesXMLMessage;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.concurrent.atomic.AtomicBoolean;
import org.apache.beam.sdk.io.UnboundedSource;
import org.apache.beam.sdk.io.UnboundedSource.UnboundedReader;
import org.apache.beam.sdk.io.solace.broker.MessageReceiver;
import org.apache.beam.sdk.io.solace.broker.SempClient;
import org.apache.beam.sdk.io.solace.broker.SessionService;
import org.apache.beam.sdk.transforms.windowing.BoundedWindow;
import org.apache.beam.vendor.guava.v32_1_2_jre.com.google.common.annotations.VisibleForTesting;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.joda.time.Instant;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** Unbounded Reader to read messages from a Solace Router. */
@VisibleForTesting
class UnboundedSolaceReader<T> extends UnboundedReader<T> {

  private static final Logger LOG = LoggerFactory.getLogger(UnboundedSolaceReader.class);
  private final UnboundedSolaceSource<T> currentSource;
  private final WatermarkPolicy<T> watermarkPolicy;
  private final SempClient sempClient;
  private @Nullable BytesXMLMessage solaceOriginalRecord;
  private @Nullable T solaceMappedRecord;
  private @Nullable MessageReceiver messageReceiver;
  private @Nullable SessionService sessionService;
  AtomicBoolean active = new AtomicBoolean(true);

  /**
   * Queue to place advanced messages before {@link #getCheckpointMark()} be called non-concurrent
   * queue, should only be accessed by the reader thread A given {@link UnboundedReader} object will
   * only be accessed by a single thread at once.
   */
  private final java.util.Queue<BytesXMLMessage> elementsToCheckpoint = new ArrayDeque<>();

  public UnboundedSolaceReader(UnboundedSolaceSource<T> currentSource) {
    this.currentSource = currentSource;
    this.watermarkPolicy = WatermarkPolicy.create(currentSource.getTimestampFn());
    this.sessionService = currentSource.getSessionServiceFactory().create();
    this.sempClient = currentSource.getSempClientFactory().create();
  }

  @Override
  public boolean start() {
    populateSession();
    populateMessageConsumer();
    return advance();
  }

  public void populateSession() {
    if (sessionService == null) {
      sessionService = getCurrentSource().getSessionServiceFactory().create();
    }
    if (sessionService.isClosed()) {
      checkNotNull(sessionService).connect();
    }
  }

  private void populateMessageConsumer() {
    if (messageReceiver == null) {
      messageReceiver = checkNotNull(sessionService).createReceiver();
      messageReceiver.start();
    }
    MessageReceiver receiver = checkNotNull(messageReceiver);
    if (receiver.isClosed()) {
      receiver.start();
    }
  }

  @Override
  public boolean advance() {
    BytesXMLMessage receivedXmlMessage;
    try {
      receivedXmlMessage = checkNotNull(messageReceiver).receive();
    } catch (IOException e) {
      LOG.warn("SolaceIO.Read: Exception when pulling messages from the broker.", e);
      return false;
    }

    if (receivedXmlMessage == null) {
      return false;
    }
    elementsToCheckpoint.add(receivedXmlMessage);
    solaceOriginalRecord = receivedXmlMessage;
    solaceMappedRecord = getCurrentSource().getParseFn().apply(receivedXmlMessage);
    watermarkPolicy.update(solaceMappedRecord);
    return true;
  }

  @Override
  public void close() {
    active.set(false);
    checkNotNull(sessionService).close();
  }

  @Override
  public Instant getWatermark() {
    // should be only used by a test receiver
    if (checkNotNull(messageReceiver).isEOF()) {
      return BoundedWindow.TIMESTAMP_MAX_VALUE;
    }
    return watermarkPolicy.getWatermark();
  }

  @Override
  public UnboundedSource.CheckpointMark getCheckpointMark() {
    List<BytesXMLMessage> ackQueue = new ArrayList<>();
    while (!elementsToCheckpoint.isEmpty()) {
      BytesXMLMessage msg = elementsToCheckpoint.poll();
      if (msg != null) {
        ackQueue.add(msg);
      }
    }
    return new SolaceCheckpointMark(active, ackQueue);
  }

  @Override
  public T getCurrent() throws NoSuchElementException {
    if (solaceMappedRecord == null) {
      throw new NoSuchElementException();
    }
    return solaceMappedRecord;
  }

  @Override
  public byte[] getCurrentRecordId() throws NoSuchElementException {
    if (solaceOriginalRecord == null) {
      throw new NoSuchElementException();
    }
    if (solaceOriginalRecord.getApplicationMessageId() != null) {
      return checkNotNull(solaceOriginalRecord)
          .getApplicationMessageId()
          .getBytes(StandardCharsets.UTF_8);
    } else {
      return checkNotNull(solaceOriginalRecord)
          .getReplicationGroupMessageId()
          .toString()
          .getBytes(StandardCharsets.UTF_8);
    }
  }

  @Override
  public UnboundedSolaceSource<T> getCurrentSource() {
    return currentSource;
  }

  @Override
  public Instant getCurrentTimestamp() throws NoSuchElementException {
    if (getCurrent() == null) {
      throw new NoSuchElementException();
    }
    return currentSource.getTimestampFn().apply(getCurrent());
  }

  @Override
  public long getTotalBacklogBytes() {
    try {
      return sempClient.getBacklogBytes(currentSource.getQueue().getName());
    } catch (IOException e) {
      LOG.warn("SolaceIO.Read: Could not query backlog bytes. Returning BACKLOG_UNKNOWN", e);
      return BACKLOG_UNKNOWN;
    }
  }
}

==> apache-beam-2.59.0-source-release.zip: beam-2.59.0/sdks/java/io/solace/src/main/java/org/apache/beam/sdk/io/solace/read/UnboundedSolaceReader.java <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.beam.sdk.io.solace.read;

import static org.apache.beam.vendor.guava.v32_1_2_jre.com.google.common.base.Preconditions.checkNotNull;

import com.solacesystems.jcsmp.BytesXMLMessage;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.concurrent.atomic.AtomicBoolean;
import org.apache.beam.sdk.io.UnboundedSource;
import org.apache.beam.sdk.io.UnboundedSource.UnboundedReader;
import org.apache.beam.sdk.io.solace.broker.MessageReceiver;
import org.apache.beam.sdk.io.solace.broker.SempClient;
import org.apache.beam.sdk.io.solace.broker.SessionService;
import org.apache.beam.sdk.transforms.windowing.BoundedWindow;
import org.apache.beam.vendor.guava.v32_1_2_jre.com.google.common.annotations.VisibleForTesting;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.joda.time.Instant;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** Unbounded Reader to read messages from a Solace Router. */
@VisibleForTesting
class UnboundedSolaceReader<T> extends UnboundedReader<T> {

  private static final Logger LOG = LoggerFactory.getLogger(UnboundedSolaceReader.class);
  private final UnboundedSolaceSource<T> currentSource;
  private final WatermarkPolicy<T> watermarkPolicy;
  private final SempClient sempClient;
  private @Nullable BytesXMLMessage solaceOriginalRecord;
  private @Nullable T solaceMappedRecord;
  private @Nullable MessageReceiver messageReceiver;
  private @Nullable SessionService sessionService;
  AtomicBoolean active = new AtomicBoolean(true);

  /**
   * Queue to place advanced messages before {@link #getCheckpointMark()} be called non-concurrent
   * queue, should only be accessed by the reader thread A given {@link UnboundedReader} object will
   * only be accessed by a single thread at once.
   */
  private final java.util.Queue<BytesXMLMessage> elementsToCheckpoint = new ArrayDeque<>();

  public UnboundedSolaceReader(UnboundedSolaceSource<T> currentSource) {
    this.currentSource = currentSource;
    this.watermarkPolicy =
        WatermarkPolicy.create(
            currentSource.getTimestampFn(), currentSource.getWatermarkIdleDurationThreshold());
    this.sessionService = currentSource.getSessionServiceFactory().create();
    this.sempClient = currentSource.getSempClientFactory().create();
  }

  @Override
  public boolean start() {
    populateSession();
    populateMessageConsumer();
    return advance();
  }

  public void populateSession() {
    if (sessionService == null) {
      sessionService = getCurrentSource().getSessionServiceFactory().create();
    }
    if (sessionService.isClosed()) {
      checkNotNull(sessionService).connect();
    }
  }

  private void populateMessageConsumer() {
    if (messageReceiver == null) {
      messageReceiver = checkNotNull(sessionService).createReceiver();
      messageReceiver.start();
    }
    MessageReceiver receiver = checkNotNull(messageReceiver);
    if (receiver.isClosed()) {
      receiver.start();
    }
  }

  @Override
  public boolean advance() {
    BytesXMLMessage receivedXmlMessage;
    try {
      receivedXmlMessage = checkNotNull(messageReceiver).receive();
    } catch (IOException e) {
      LOG.warn("SolaceIO.Read: Exception when pulling messages from the broker.", e);
      return false;
    }

    if (receivedXmlMessage == null) {
      return false;
    }
    elementsToCheckpoint.add(receivedXmlMessage);
    solaceOriginalRecord = receivedXmlMessage;
    solaceMappedRecord = getCurrentSource().getParseFn().apply(receivedXmlMessage);
    watermarkPolicy.update(solaceMappedRecord);
    return true;
  }

  @Override
  public void close() {
    active.set(false);
    checkNotNull(sessionService).close();
  }

  @Override
  public Instant getWatermark() {
    // should be only used by a test receiver
    if (checkNotNull(messageReceiver).isEOF()) {
      return BoundedWindow.TIMESTAMP_MAX_VALUE;
    }
    return watermarkPolicy.getWatermark();
  }

  @Override
  public UnboundedSource.CheckpointMark getCheckpointMark() {
    List<BytesXMLMessage> ackQueue = new ArrayList<>();
    while (!elementsToCheckpoint.isEmpty()) {
      BytesXMLMessage msg = elementsToCheckpoint.poll();
      if (msg != null) {
        ackQueue.add(msg);
      }
    }
    return new SolaceCheckpointMark(active, ackQueue);
  }

  @Override
  public T getCurrent() throws NoSuchElementException {
    if (solaceMappedRecord == null) {
      throw new NoSuchElementException();
    }
    return solaceMappedRecord;
  }

  @Override
  public byte[] getCurrentRecordId() throws NoSuchElementException {
    if (solaceOriginalRecord == null) {
      throw new NoSuchElementException();
    }
    if (solaceOriginalRecord.getApplicationMessageId() != null) {
      return checkNotNull(solaceOriginalRecord)
          .getApplicationMessageId()
          .getBytes(StandardCharsets.UTF_8);
    } else {
      return checkNotNull(solaceOriginalRecord)
          .getReplicationGroupMessageId()
          .toString()
          .getBytes(StandardCharsets.UTF_8);
    }
  }

  @Override
  public UnboundedSolaceSource<T> getCurrentSource() {
    return currentSource;
  }

  @Override
  public Instant getCurrentTimestamp() throws NoSuchElementException {
    if (getCurrent() == null) {
      throw new NoSuchElementException();
    }
    return currentSource.getTimestampFn().apply(getCurrent());
  }

  @Override
  public long getTotalBacklogBytes() {
    try {
      return sempClient.getBacklogBytes(currentSource.getQueue().getName());
    } catch (IOException e) {
      LOG.warn("SolaceIO.Read: Could not query backlog bytes. Returning BACKLOG_UNKNOWN", e);
      return BACKLOG_UNKNOWN;
    }
  }
}
